#!/bin/bash

# Copyright (c) 2022 xFusion Digital Technologies Co., Ltd. All rights reserved.
# Description: build hotpatch rpm pakage
# Author: wuqiong
# Create: 2022-01-22

CUR_DIR=$(cd "$(dirname "${BASH_SOURCE-$0}")"; pwd)
cd ${CUR_DIR}

#rpm info
PATCH_KERNEL_VERSION=
PATCH_KERNEL_VERSION_WITHOUT_ARCH=
PATCH_LOCATION=
PATCH_NAME=
PATCH_FULL_NAME=
PATCH_REQUIRES=
PATCH_OBSOLETES=
PATCH_DESC=
PATCH_PACK_ENV_ROOT=
PATCH_ARCH=`arch`

PATCH_MODULE_SRC=
PATCH_MODULE_MAKEFILE=
PATCH_BUILD_JOBS=16

#You can pre-define the followings via envionment variables
: ${PATCH_VERSION:=1.0}
: ${PATCH_RELEASE:=1}
: ${PATCH_TIME:=`date +%s`}

#error code
readonly PATCH_BUILD_ERROR_SUCCESS=0
readonly PATCH_BUILD_ERROR_INVALID_ARGS=1
readonly PATCH_BUILD_ERROR_BUILD_KPATCH=2
readonly PATCH_BUILD_ERROR_PARSE_KPATCH=3
readonly PATCH_BUILD_FIND_DEPS=4
readonly PATCH_BUILD_NAME_DUPLICATE=5
readonly PATCH_BUILD_ERROR_BUILD_RPM=6

#klp requires
PATCH_KLP_REQUIRES=""

#added file name
readonly PATCH_KLP_CHANGES="klp-changes"
readonly PATCH_KLP_INFO="klp-info"

PATCH_BUILD_TMP_DIR="${CUR_DIR}/tmp"
PATCH_PACKAGE_PATH=

source "${CUR_DIR}/parse_klp.sh"

function initialize() {
    if [ -d ${PATCH_BUILD_TMP_DIR} ]; then
        rm -rf ${PATCH_BUILD_TMP_DIR}
    fi

    echo "Create tmp directory \"${PATCH_BUILD_TMP_DIR}\" for build."
    mkdir -p ${PATCH_BUILD_TMP_DIR}
    mkdir -p ~/rpmbuild/{RPMS,SRPMS,BUILD,BUILDROOT,SOURCES,SPECS}
}

function die() {
    local exit_code=$1
    exit ${exit_code}
}

function usage() {
cat <<EOF
    Usage:
        build.sh [OPTIONS] -e pack_env_root -l source-patch-path -n patch_name -k kernel-version -d description

    Options:
        -e, --pack-env-root
                Specifies the root path of the packing environment, which contains the Fusion compile 
                env and the released hotpatches organized as below:
                    <pack-env-root>/
                        <kernel-version>/Fusion_compile_env.tar.gz
                        <kernel-version>/rpms/klp_*.rpm
        -m, --module-src
                Specifies the out-of-tree module src path for making out-of-tree module hotpatch.
        -f, --makefile
                Specifies the out-of-tree module makefile for making out-of-tree module hotpatch.
        -j, --jobs
                Specifies the number of jobs to run simultaneously while performing builds.
        -l, --patch-location
                The patch of the source code which is used to make hotpatch.
        -n, --patch-name
                Specifies the patch name.
        -k, --kernel-version
                Specifies the kernel version for making the hotpatch.
        -r, --requires
                Specifies the alternative rpm requirements.
        -o, --obsoletes
                Specifies the obsoletes that the hotpatch will replace.
        -d, --description
                Specifies the description for the hotpatch.
        -h, --help
                help info.
EOF
}

function pre_check_args() {
    if [ -z ${PATCH_LOCATION} ]; then
        echo "Please specify the patch path!"
        die ${PATCH_BUILD_ERROR_INVALID_ARGS}
    elif [ ! -e ${PATCH_LOCATION} ]; then
        echo "The patch \"${PATCH_LOCATION}\" doesn't exist!"
        die ${PATCH_BUILD_ERROR_INVALID_ARGS}
    fi
    PATCH_LOCATION=`readlink -e ${PATCH_LOCATION}`

    if [ -z ${PATCH_PACK_ENV_ROOT} ]; then
        echo "Please specify the pack env root path"
        die ${PATCH_BUILD_ERROR_INVALID_ARGS}
    elif [ ! -e ${PATCH_PACK_ENV_ROOT} ]; then
        echo "The pack env root path \"${PATCH_PACK_ENV_ROOT}\" doesn't exist!"
        die ${PATCH_BUILD_ERROR_INVALID_ARGS}
    fi
    PATCH_PACK_ENV_ROOT=`readlink -e ${PATCH_PACK_ENV_ROOT}`

    if [ ! -z ${PATCH_MODULE_SRC} ]; then
        if [ ! -e ${PATCH_MODULE_SRC} ]; then
            echo "The module source path \"${PATCH_MODULE_SRC}\" doesn't exist!"
            die ${PATCH_BUILD_ERROR_INVALID_ARGS}
        fi
        PATCH_MODULE_SRC=`readlink -e ${PATCH_MODULE_SRC}`
    fi

    if [ ! -z ${PATCH_MODULE_MAKEFILE} ]; then
        if [ ! -e ${PATCH_MODULE_MAKEFILE} ]; then
            echo "The module makefile \"${PATCH_MODULE_MAKEFILE}\" doesn't exist!"
            die ${PATCH_BUILD_ERROR_INVALID_ARGS}
        fi
        PATCH_MODULE_MAKEFILE=`readlink -e ${PATCH_MODULE_MAKEFILE}`
    fi

    if [ -z ${PATCH_NAME} ]; then
        echo "Please specify the patch name!"
        die ${PATCH_BUILD_ERROR_INVALID_ARGS}
    fi
    if [ "${PATCH_NAME:0:4}" != "klp_" ]; then
        echo "Patch name should start with \"klp_\"!"
        die ${PATCH_BUILD_ERROR_INVALID_ARGS}
    fi

    if [ -z ${PATCH_KERNEL_VERSION} ]; then
        echo "Please specify the kernel version for the patch!"
       die ${PATCH_BUILD_ERROR_INVALID_ARGS}
    fi
    PATCH_KERNEL_VERSION_WITHOUT_ARCH=${PATCH_KERNEL_VERSION%.${PATCH_ARCH}}
    PATCH_FULL_NAME=${PATCH_NAME}-${PATCH_KERNEL_VERSION_WITHOUT_ARCH}

    PATCH_PACK_ENV_ROOT="${PATCH_PACK_ENV_ROOT}/${PATCH_KERNEL_VERSION}"
    if [ ! -e ${PATCH_PACK_ENV_ROOT} ]; then
        echo "The pack env kernel path \"${PATCH_PACK_ENV_ROOT}\" doesn't exist!"
        die ${PATCH_BUILD_ERROR_INVALID_ARGS}
    fi
    if [ ! -e "${PATCH_PACK_ENV_ROOT}/rpms" ]; then
        echo "${PATCH_PACK_ENV_ROOT}/rpms, which is used to cache released hotpatches, doesn't exist!"
        die ${PATCH_BUILD_ERROR_INVALID_ARGS}
    fi

    local match=`expr match "${PATCH_REQUIRES}" "kpatch-runtime"`
    if [ "${match}" == "0" ]; then
        PATCH_REQUIRES="${PATCH_REQUIRES} kpatch-runtime"
    fi
}

function parse_args() {
    local input_args=`getopt -o he:m:f:j:l:n:k:r:o:d: \
        --long help,pack-env-root:,module-src:,makefile:,jobs:,patch-location:,patch-name:,kernel-version:,requires:,obsoletes:,description: -- "$@"`

    if [ $? != 0 ]; then
        echo "Terminating..." >&2;
        exit 1;
    fi

    eval set -- "${input_args}"
    while true; do
        case $1 in
            -h | --help)
                usage;
                exit 0
                ;;
            -e | --pack-env-root)
                PATCH_PACK_ENV_ROOT=$2
                shift 2
                ;;
            -m | --module-src)
                PATCH_MODULE_SRC=$2
                shift 2
                ;;
            -f | --makefile)
                PATCH_MODULE_MAKEFILE=$2
                shift 2
                ;;
            -j | --jobs)
                if [ $2 != 0 ]; then
                    PATCH_BUILD_JOBS=$2
                fi
                shift 2
                ;;
            -l | --patch-location)
                PATCH_LOCATION=$2
                shift 2
                ;;
            -n | --patch-name)
                 PATCH_NAME=$2
                 shift 2
                 ;;
            -k | --kernel-version)
                PATCH_KERNEL_VERSION=$2
                shift 2
                ;;
            -r | --requires)
                PATCH_REQUIRES=$2
                shift 2
                ;;
            -o | --obsoletes)
                PATCH_OBSOLETES=$2
                shift 2
                ;;
            -d | --description)
                PATCH_DESC=$2
                shift 2
                ;;
            --)
                shift;
                break
                ;;
            *)
                echo "Invalid arguments!";
                die ${PATCH_BUILD_ERROR_INVALID_ARGS}
                ;;
        esac
    done

    pre_check_args
}

declare -a g_require_rpm_name_array
declare -a g_require_rpm_build_time_array
declare -a g_require_rpm_pkg_index_array
declare -a g_require_klp_name_array
declare -a g_require_rpm_changes_str_array
declare -a g_require_index_del_array
g_history_package_count=0
g_current_package_index=0

function rebuild_require_arrays() {
    local require_array_len=${#g_require_rpm_name_array}
    local is_found=0

    if [ ${#g_require_index_del_array[@]} -gt 0 ]; then
        local -a tmp_g_require_rpm_name_array
        local -a tmp_g_require_rpm_build_time_array
        local -a tmp_g_require_rpm_pkg_index_array
        local -a tmp_g_require_rpm_changes_str_array
        local -a tmp_g_require_klp_name_array

        for ((index=0; index<${require_array_len}; index++))
        do
            is_found=0
            for ((jj=0; jj<${#g_require_index_del_array[@]}; jj++))
            do
                if [ "${g_require_index_del_array[jj]}" = "${index}" ]; then
                    is_found=1
                    break
                fi
            done

            if [ ${is_found} -eq 0 ]; then
                tmp_g_require_rpm_name_array=(${tmp_g_require_rpm_name_array[@]} ${g_require_rpm_name_array[${index}]})
                tmp_g_require_rpm_build_time_array=(${tmp_g_require_rpm_build_time_array[@]} ${g_require_rpm_build_time_array[${index}]})
                tmp_g_require_rpm_pkg_index_array=(${tmp_g_require_rpm_pkg_index_array[@]} ${g_require_rpm_pkg_index_array[${index}]})
                tmp_g_require_klp_name_array=(${tmp_g_require_klp_name_array[@]} ${g_require_klp_name_array[${index}]})
                tmp_g_require_rpm_changes_str_array=(${tmp_g_require_rpm_changes_str_array[@]} ${g_require_rpm_changes_str_array[${index}]})
            fi
        done

        g_require_rpm_name_array=${tmp_g_require_rpm_name_array[@]}
        g_require_rpm_build_time_array=${tmp_g_require_rpm_build_time_array[@]}
        g_require_rpm_pkg_index_array=${tmp_g_require_rpm_pkg_index_array[@]}
        g_require_klp_name_array=${tmp_g_require_klp_name_array[@]}
        g_require_rpm_changes_str_array=${tmp_g_require_rpm_changes_str_array[@]}
    fi
}

function initialize_find_dependencies() {
    echo "initialize_find_dependencies:"

    g_require_rpm_name_array=()
    g_require_rpm_build_time_array=()
    g_require_klp_name_array=()
    g_require_rpm_changes_str_array=()
    g_require_index_del_array=()
    g_history_package_count=0
    g_current_package_index=0
}

function finalize_find_dependencies() {
    echo "finalize_find_dependencies:"
    echo "--------------------------------------------------------------------------------"
    echo "The patch requires rpm: \"${g_require_rpm_name_array[@]}\""
    for require_rpm in ${g_require_rpm_name_array[@]}
    do
        PATCH_REQUIRES="${PATCH_REQUIRES} ${require_rpm}"
    done
    echo "${PATCH_REQUIRES}"

    echo "--------------------------------------------------------------------------------"
    echo "The patch requires klp: \"${g_require_klp_name_array[@]}\""
    for require_klp in ${g_require_klp_name_array[@]}
    do
        if [ ! -z "${PATCH_KLP_REQUIRES}" ]; then
            PATCH_KLP_REQUIRES="${PATCH_KLP_REQUIRES} "
        fi
        PATCH_KLP_REQUIRES="${PATCH_KLP_REQUIRES}${require_klp}"
    done
    echo "${PATCH_KLP_REQUIRES}"
    echo "--------------------------------------------------------------------------------"
}

function gen_klp_changes_string() {
    local changes_string=""
    for mod in $(echo ${!KLP_MOD_FUNCS_DICT[*]})
    do
        changes_string="${changes_string}${mod}:${KLP_MOD_FUNCS_DICT[${mod}]}\n"
    done

    echo -e "${changes_string}"
}

function gen_klp_modules_string() {
    local modules_string=""
    for mod in $(echo ${!KLP_MOD_FUNCS_DICT[*]})
    do
        if [ ! -z "${modules_string}" ]; then
            modules_string="${modules_string} "
        fi
        modules_string="${modules_string}${mod}"
    done

    echo -e "${modules_string}"
}

function is_change_conflict() {
    local klp_changes_str_a=$1
    local klp_changes_str_b=$2

    echo "is_change_conflict:"
    local -A klp_changes_dict_a
    echo "--------------------------------------------------------------------------------"
    for mod in `echo "${klp_changes_str_a}" | awk -F ":" '{print $1}'`
    do
        local change_funcs_a=`echo "${klp_changes_str_a}" | grep "${mod}:" |awk -F ":" '{print $2}'`
        klp_changes_dict_a[${mod}]="${change_funcs_a[@]}"
        echo "a-${mod} : ${klp_changes_dict_a[${mod}]}"
    done

    local -A klp_changes_dict_b
    for mod in `echo "${klp_changes_str_b}" | awk -F ":" '{print $1}'`
    do
        local change_funcs_b=`echo "${klp_changes_str_b}" | grep "${mod}:" |awk -F ":" '{print $2}'`
        klp_changes_dict_b[${mod}]="${change_funcs_b[@]}"
        echo "b-${mod} : ${klp_changes_dict_b[${mod}]}"
    done
    echo "--------------------------------------------------------------------------------"

    for mod in $(echo ${!klp_changes_dict_a[*]})
    do
        local change_funcs_a=${klp_changes_dict_a[${mod}]}
        local change_funcs_b=${klp_changes_dict_b[${mod}]}

        echo "${mod}:${change_funcs_a[@]}"
        echo "${mod}:${change_funcs_b[@]}"
        for change_func_a in ${change_funcs_a[@]}
        do
            for change_func_b in ${change_funcs_b[@]}
            do
                if [ "${change_func_a}" = "${change_func_b}" ]; then
                    echo "function change \"${change_func_a}@${mod}\" overlap!"
                    return 1
                fi
            done
        done
    done

    return 0
}

function check_one_rpm_conflict() {
    local repo_rpm=$1

    local repo_rpm_name=`basename ${repo_rpm}`
    repo_rpm_name="${repo_rpm_name%${PATCH_KERNEL_VERSION_WITHOUT_ARCH}*}${PATCH_KERNEL_VERSION_WITHOUT_ARCH}"
    local repo_patch_name=${repo_rpm_name%-${PATCH_KERNEL_VERSION_WITHOUT_ARCH}*}

    if [ "${PATCH_NAME}" = "${repo_patch_name}" ]; then
        echo "The patch name duplicates with released patch \"${repo_rpm_name}\"."
        die ${PATCH_BUILD_NAME_DUPLICATE}
    fi

    cd ${PATCH_BUILD_TMP_DIR}
    rpm2cpio ${repo_rpm} | cpio -div
    cd ${CUR_DIR}

    local install_path="${PATCH_BUILD_TMP_DIR}/usr/lib/modules/${PATCH_KERNEL_VERSION}/kernel-hotpatch/${repo_patch_name}"
    local rpm_build_time=`cat ${install_path}/${PATCH_KLP_INFO} | grep compiletime | awk -F "=" '{print $2}'`
    local rpm_pkg_index=`cat ${install_path}/${PATCH_KLP_INFO} | grep pkgindex | awk -F "=" '{print $2}'`
    if [ -z "${rpm_build_time}" ] || [ ${rpm_build_time} -ge ${PATCH_TIME} ]; then
        echo "Build time \"${rpm_build_time}\" of \"${rpm}\" is greater than current build time \"${PATCH_TIME}\"."
        die ${PATCH_BUILD_FIND_DEPS}
    fi

    #check if the rpm conflicts with current klp
    echo "Checking if the rpm conflicts with the klp..."

    local repo_rpm_changes_str=`cat ${install_path}/${PATCH_KLP_CHANGES}`

    echo "--------------------------------------------------------------------------------"
    echo "\"${repo_rpm_name}\" changes:"
    echo "${repo_rpm_changes_str}"
    echo "--------------------------------------------------------------------------------"

    local klp_changes_str=$(gen_klp_changes_string)
    echo "--------------------------------------------------------------------------------"
    echo "klp changes:"
    echo "${klp_changes_str}"
    echo "--------------------------------------------------------------------------------"

    is_change_conflict "${repo_rpm_changes_str}" "${klp_changes_str}"
    local is_conflict=$?
    if [ ${is_conflict} -eq 0 ]; then
        echo "The rpm doesn't conflict with the klp."
        return 0
    fi

    #check if conflict with the candidates requirements
    local is_new_require=1
    local require_array_len=${#g_require_rpm_name_array[@]}
    echo "Continue to check if the rpm conflicts with the candidate required ${require_array_len} rpms."
    for ((index=0; index<${require_array_len}; index++))
    do
        local require_rpm_changes_str="${g_require_rpm_changes_str_array[${index}]}"
        local require_rpm_name=${g_require_rpm_name_array[${index}]}
        is_change_conflict "${repo_rpm_changes_str}" "${require_rpm_changes_str}"
        is_conflict=$?

        if [ ${is_conflict} -gt 0 ]; then
            echo "The rpm conflicts with the candidate required rpm \"${require_rpm_name}\""
            #if the candidate require is newer than the rpm
            require_build_time=${g_require_rpm_build_time_array[${index}]}
            if [ ${require_build_time} -gt ${rpm_build_time} ]; then
                echo "The rpm is older than the candidate required rpm \"${require_rpm_name}. Ignore the rpm."
                is_new_require=0
            else
                #if candidate require is older than the rpm
                echo "The rpm is newer than the candidate required rpm \"${require_rpm_name}. Remove the candidate required rpm."
                g_require_index_del_array[${#g_require_index_del_array[@]}]=${index}
            fi
        fi
    done

    #delete old requires
    rebuild_require_arrays

    #add new require
    require_array_len=${#g_require_rpm_name_array[@]}
    if [ ${is_new_require} -gt 0 ]; then
        echo "Added rpm \"${repo_rpm_name}\" as a candidate required rpm."
        g_require_rpm_name_array[${require_array_len}]="${repo_rpm_name}"
        g_require_rpm_build_time_array[${require_array_len}]="${rpm_build_time}"
        g_require_rpm_pkg_index_array[${require_array_len}]="${rpm_pkg_index}"
        g_require_klp_name_array[${require_array_len}]="${repo_patch_name}"
        g_require_rpm_changes_str_array[${require_array_len}]="${repo_rpm_changes_str}"
    fi

    return ${is_new_require}
}

function find_dependencies() {
    initialize_find_dependencies
    local repo_rpms=`ls ${PATCH_PACK_ENV_ROOT}/rpms/klp_*-${PATCH_KERNEL_VERSION_WITHOUT_ARCH}-*.rpm 2>/dev/null`

    if [ -z "${repo_rpms}" ]; then
        g_history_package_count=0
    else
        g_history_package_count=`echo "${repo_rpms[@]}" | wc -l`
    fi
    ((g_current_package_index=$g_history_package_count+1))

    echo "history package count=$g_history_package_count"
    echo "current package index=$g_current_package_index"

    for repo_rpm in ${repo_rpms[@]}
    do
        echo "================================================================================="
        echo "Checking existing rpm \"${repo_rpm}\""
        #check one rpm
        repo_rpm_path=`readlink -e ${repo_rpm}`
        check_one_rpm_conflict ${repo_rpm_path}
        echo "================================================================================="
    done
    finalize_find_dependencies
}

function create_installation_archive() {
    local package_full_name=$1
    mkdir ${PATCH_BUILD_TMP_DIR}/${package_full_name}
    cp -a ${PATCH_PACKAGE_PATH} ${PATCH_BUILD_TMP_DIR}/${package_full_name}/

    #create klp-info
    touch ${PATCH_BUILD_TMP_DIR}/${package_full_name}/${PATCH_KLP_INFO}
    echo "compiletime=${PATCH_TIME}" >> ${PATCH_BUILD_TMP_DIR}/${package_full_name}/${PATCH_KLP_INFO}
    echo "pkgindex=${g_current_package_index}" >> ${PATCH_BUILD_TMP_DIR}/${package_full_name}/${PATCH_KLP_INFO}
    echo "requires=${PATCH_KLP_REQUIRES}" >> ${PATCH_BUILD_TMP_DIR}/${package_full_name}/${PATCH_KLP_INFO}

    #create klp-changes
    touch ${PATCH_BUILD_TMP_DIR}/${package_full_name}/${PATCH_KLP_CHANGES}
    local klp_changes_str=$(gen_klp_changes_string)
    echo "${klp_changes_str}" > ${PATCH_BUILD_TMP_DIR}/${package_full_name}/${PATCH_KLP_CHANGES}


    eval "tar jcvf ${PATCH_BUILD_TMP_DIR}/${package_full_name}.tar.bz2 -C ${PATCH_BUILD_TMP_DIR} ${package_full_name}"
}

function gen_rpm_spec() {
    local tmp_spec=${PATCH_BUILD_TMP_DIR}/${PATCH_FULL_NAME}.spec

    cp -f ${CUR_DIR}/rpms/spec/fusionos_kernel_patch.spec ${tmp_spec}

    sed -i "s%define patch_version.*%define patch_version    ${PATCH_VERSION}%" ${tmp_spec}
    sed -i "s%define patch_release.*%define patch_release    ${PATCH_RELEASE}%" ${tmp_spec}
    sed -i "s%define kernel_version.*%define kernel_version    ${PATCH_KERNEL_VERSION}%" ${tmp_spec}
    sed -i "s%define patch_name.*%define patch_name    ${PATCH_NAME}%" ${tmp_spec}
    sed -i "s%define patch_fullname.*%define patch_fullname    ${PATCH_FULL_NAME}%" ${tmp_spec}

    local patch_pkg_name=`basename ${PATCH_PACKAGE_PATH}`
    sed -i "s%define patch_package_name.*%define patch_package_name    ${patch_pkg_name}%" ${tmp_spec}

    if [ ! -z "${PATCH_DESC}" ]; then
        sed -i "s%define patch_desc.*%define patch_desc    ${PATCH_DESC}%" ${tmp_spec}
    fi

    if [ ! -z "${PATCH_KLP_REQUIRES}" ]; then
        sed -i "s%define patch_klp_requires.*%define patch_klp_requires    ${PATCH_KLP_REQUIRES}%" ${tmp_spec}
    fi

    local patch_mod_requires=$(gen_klp_modules_string)
    if [ ! -z "${patch_mod_requires}" ]; then
        sed -i "s%define patch_mod_requires.*%define patch_mod_requires    ${patch_mod_requires}%" ${tmp_spec}
    fi
}

function build_rpm_package() {
    local package_full_name="${PATCH_FULL_NAME}-${PATCH_VERSION}"

    gen_rpm_spec
    create_installation_archive ${package_full_name}

    rm -rf ~/rpmbuild/SOURCES/${PATCH_FULL_NAME}* ~/rpmbuild/RPMS/${PATCH_ARCH}/${PATCH_FULL_NAME}* ~/rpmbuild/SRPMS/${PATCH_FULL_NAME}*
    rm -rf ~/rpmbuild/BUILD/${PATCH_FULL_NAME}*
    mv ${PATCH_BUILD_TMP_DIR}/${package_full_name}.tar.bz2 ~/rpmbuild/SOURCES/

    local rpmbuild_param="${PATCH_BUILD_TMP_DIR}/${PATCH_FULL_NAME}.spec"
    if [ ! -z "${PATCH_REQUIRES}" ]; then
        rpmbuild_param="${rpmbuild_param} -D 'patch_requires ${PATCH_REQUIRES}'"
    fi

    if [ ! -z "${PATCH_OBSOLETES}" ]; then
        rpmbuild_param="${rpmbuild_param} -D 'patch_obsoletes ${PATCH_OBSOLETES}'"
    fi

    eval "rpmbuild -ba ${rpmbuild_param} --nodebuginfo"
    mkdir -p ${CUR_DIR}/rpms/dist
    cp -a ~/rpmbuild/RPMS/${PATCH_ARCH}/${PATCH_FULL_NAME}* ${CUR_DIR}/rpms/dist
}

function parse_ko() {
    tar -zxf ${PATCH_PACKAGE_PATH} -C ${PATCH_BUILD_TMP_DIR}
    klp_parse "${PATCH_BUILD_TMP_DIR}/${PATCH_NAME}/${PATCH_NAME}.ko"
    if [ $? != 0 ]; then
        echo "Failed to parse kpatch module."
        die ${PATCH_BUILD_ERROR_PARSE_KPATCH}
    fi
}

function build_ko() {
    local comple_env_name="FusionOS_compile_env"

    #check if the name already been used.
    PATCH_PACKAGE_PATH=${PATCH_PACK_ENV_ROOT}/${comple_env_name}/opt/patch_workspace/hotpatch/${PATCH_NAME}.tar.gz
    if [ -e ${PATCH_PACKAGE_PATH} ]; then
        echo "The patch name has already been used, please use a different one!"
        die ${PATCH_BUILD_ERROR_BUILD_KPATCH}
    fi

    if [ ! -e "${PATCH_PACK_ENV_ROOT}/${comple_env_name}" ]; then
        local comile_env="${PATCH_PACK_ENV_ROOT}/${comple_env_name}.tar.gz"
        echo "Extracting \"${comile_env}\" to \"${PATCH_PACK_ENV_ROOT}\"..."
        tar -zxf ${comile_env} -C ${PATCH_PACK_ENV_ROOT}/
    fi

    local make_hotpatch_args="-j ${PATCH_BUILD_JOBS} -i ${PATCH_NAME#klp_}"

    #copy source patch to compile env
    diff_in_compile_env="${PATCH_PACK_ENV_ROOT}/${comple_env_name}/opt/patch_workspace/diff/"
    echo "Copying source code patch \"${PATCH_LOCATION}\" to compile env \"${diff_in_compile_env}\"..."
    mkdir -p ${diff_in_compile_env}
    cp ${PATCH_LOCATION} ${diff_in_compile_env}
    diff_file_name=`basename ${PATCH_LOCATION}`
    patch_loc_in_compile_env="/opt/patch_workspace/diff/${diff_file_name}"
    make_hotpatch_args="${make_hotpatch_args} -p ${patch_loc_in_compile_env}"

    #create tmp folder inside compile env for out-of-tree module
    local random="${RANDOM}"
    if [ -z "${random}" ]; then
        random=`date +%s`
    fi
    local module_tmp_dir_compile_env="/tmp/module.${random}"
    local module_tmp_dir="${PATCH_PACK_ENV_ROOT}/${comple_env_name}/${module_tmp_dir_compile_env}"
    if [ ! -z ${PATCH_MODULE_SRC} ]; then
        mkdir -m 700 -p ${module_tmp_dir}/${PATCH_MODULE_SRC}
        echo "Copying module source \"${PATCH_MODULE_SRC}\" to compile env \"${module_tmp_dir}/${PATCH_MODULE_SRC}\"..."
        cp -Rf ${PATCH_MODULE_SRC}/* ${module_tmp_dir}/${PATCH_MODULE_SRC}
        make_hotpatch_args="${make_hotpatch_args} -m ${module_tmp_dir_compile_env}/${PATCH_MODULE_SRC}"
    fi

    if [ ! -z ${PATCH_MODULE_MAKEFILE} ]; then
        local makefile_dir=`dirname ${PATCH_MODULE_MAKEFILE}`
        mkdir -m 700 -p ${module_tmp_dir}/${makefile_dir}
        echo "Copying module makefile \"${PATCH_MODULE_MAKEFILE}\" to compile env \"${module_tmp_dir}/${PATCH_MODULE_MAKEFILE}\"..."
        cp -a ${PATCH_MODULE_MAKEFILE} ${module_tmp_dir}/${PATCH_MODULE_MAKEFILE}
        make_hotpatch_args="${make_hotpatch_args} -f ${module_tmp_dir_compile_env}/${PATCH_MODULE_MAKEFILE}"
    fi

    cmd="source /etc/profile; cd /opt/patch_workspace/; ./make_hotpatch ${make_hotpatch_args}"
    echo "chroot ${PATCH_PACK_ENV_ROOT}/${comple_env_name} sh -c \"${cmd}\""
    eval "chroot ${PATCH_PACK_ENV_ROOT}/${comple_env_name} sh -c \"${cmd}\""

    if [ -e ${module_tmp_dir} ]; then
        echo "Removes module tmp dir \"${module_tmp_dir}\"."
        rm -rf ${module_tmp_dir}
    fi

    if [ ! -e ${PATCH_PACKAGE_PATH} ]; then
        echo "Failed to build kpatch module."
        die ${PATCH_BUILD_ERROR_BUILD_KPATCH}
    else
        echo "Succeeded to build kpatch module."
    fi
}

function build() {
    initialize
    build_ko
    parse_ko
    find_dependencies
    build_rpm_package
}

function main() {
    parse_args "$@"
    build
}

main "$@"
